########################################################################
# OSM::Tree:BASE

=encoding utf8

=head1 NAME

OSM::Tree::BASE -- BaseClass for Node Way Relation

=head1 SYNOPSIS


=head1 DESCRIPTION

BaseClass for

L<OSM::Tree::Node>

L<OSM::Tree::Way>

L<OSM::Tree::Relation>

=head1 AUTHOR

ToPeG

=head1 USING

L<OSM::Tree::TagList>

=head1 METHODS

=over 4

=cut

package OSM::Tree::BASE;
use OSM::Tree::TagList;
use strict;
use warnings;
use Carp;

sub _new
{
  my $class=shift;
  my $self={};
  $self->{TREE}=shift;
  $self->{ID}=shift;
  $self->{REF}=shift;
  $self->{IGNORE_DELETED}=shift // \do{my $t=0};
  $self->{FORCE_DELETE}=shift  // \do{my $t=0};
  $self->{TAGS}=undef;
  bless($self,$class);
  return $self;
}

sub _id_list
{
  my $self=shift;
  my $name=shift;
  my $elms=$self->{TREE}->{$name};
  my @l=keys(%$elms);
  if(${$self->{IGNORE_DELETED}})
  {
    my @ret=();
    for(@l)
    { push(@ret,$_) if($elms->{$_}->{action} ne 'delete'); }
    return @ret;
  }
  return @l;
}

=item id()

get the ID of this Object

returns an String of this ID

=cut
sub id
{ return shift->{ID}; }

=item has_tags()

return 1 if this object has tags or 0 otherwise

=cut
sub has_tags
{
  my $self=shift;
  my $case=shift || 0;
  return $self->tags($case)->has_tags(@_)
  #return exists($self->{REF}->{tag}) && scalar(%{$self->{REF}->{tag}});
}

=item tags(<CASESENITIVE>)

get an TagList-Object of this object

see L<OSM::Tree::TagList> for more

returns an TagList-Object or undefined, if no tags in this node

=cut
sub tags
{
  my $self=shift;
  my $case=shift || 0;
  if(!$self->{TAGS} || $self->{TAGS}->case() != $case)
  {
    $self->{REF}->{tag}={} unless($self->{REF}->{tag});
    $self->{TAGS}=OSM::Tree::TagList->_new($self,$self->{REF}->{tag},$case);
  }
  return $self->{TAGS};
}

=item has_tag(NAME,<CASESENITIVE>)

return 1 if this object has tag with this name or 0 otherwise

=cut
sub has_tag
{
  my $self=shift;
  my $name=shift;
  my $case=shift || 0;
  return $self->tags($case)->has_key($name);
  return 0;
}

=item tag(NAME,<CASESENITIVE>)

get the value of the named tag

see L<OSM::Tree::TagList> for more

returns an Tag-Object or undef if the named tag don't exists

=cut
sub tag
{
  my $self=shift;
  my $name=shift;
  my $case=shift || 0;
  return $self->tags($case)->tag($name);
  return wantarray?():undef;
}

=item user()

get the name of the creator of this node

returns a String of the autorname
or undef if the XML-tag ist not set

=cut
sub user
{ return shift->{REF}->{user}; }

=item action()

get the name of the creator of this node

returns a String of the autorname
or undef if the XML-tag ist not set

=cut
sub action
{ return shift->{REF}->{action}; }

sub _set_modify
{
  my $self=shift;
  $self->_set_action('modify') unless($self->action);
}

=item self_del()

remove themself from the OSM Tree

=cut
sub self_del
{
  my $self=shift;
  $self->_set_action('delete');
  if($self->force_delete())
  {
    delete($self->{TREE}->{lc($self->type())}->{$self->id()});
    $self->{TREE}={};
  }
}


sub _set_action
{ $_[0]->{REF}->{action}=$_[1]; }

=item user_id()

get the id of the creator of this node

returns a String of the autorname
or undef if the XML-tag ist not set

=cut
sub user_id
{ return shift->{REF}->{uid}; }

=item changeset()

???

=cut
sub changeset
{ return shift->{REF}->{changeset}; }

=item time()

get the timestamp of the creation of this node

returns a String of the Format: JJJJ-MM-DDTHH:MM:SS+HH:MM

like: 2009-03-30T19:38:29+01:00

or undef if the XML-tag ist not set

=cut
sub time
{ return shift->{REF}->{timestamp}; }

=item visible()

get the visibility of this node

returns a String contains "true" or "false"
or undef if the XML-tag ist not set

=cut
sub visible
{ return shift->{REF}->{visible}; }

=item type()

get the type of this Object

returns a String contains "Node", "Way" or "Relation"

=cut
sub type
{
  return '' unless(ref($_[0]));
  return (split(/::/m, ref($_[0])))[-1];
}

=item version()

get the version of this Object

returns a Integer Number > 0

=cut
sub version
{ return shift->{REF}->{version} || 1; }

#=======================================================================

=item relation_list([<types>,...])

get the relation-objects to this node/way/relation

if a list of Types ist set only that kind of relations will be returned

see L<OSM::Tree::Relation> for more

returns an relation-object list

=cut
sub relation_list
{
  my $self=shift;
  my $keys=undef;
  $keys={map{$_=>1}@_} if(@_);

  my @list=();

  for my $relation (keys(%{$self->{TREE}->{relation}}))
  {
    $relation=OSM::Tree::Relation->_new($self->{TREE},$relation,$self->{IGNORE_DELETED},$self->{FORCE_DELETE});
    next unless($relation);
    for my $member ($relation->member_list())
    {
      next if($keys && !$keys->{$relation->kind()});
      if(defined($member) && $member->type() eq 'way' && $member->child()->id() eq $self->{ID})
      {
        push(@list,$relation);
        last;
      }
    }
  }
  return @list;
}

=item relation_id_list()

get the relation-ids to this way

returns a relationID list

=cut
sub relation_id_list
{
  my $self=shift;
  my @list=map{$_->id()}$self->relation_list();
  return @list;
}

=item relation_list_length()

get the length the list of the relations

returns the length of the realtion-list

=cut
sub relation_list_length
{ return scalar(shift->relation_list()); }

=item force_delete(<1/0>)

If i's set the nodes/ways/relation are deleted from the Stack.
If it's not set the tag "action='delete'" is set

It's set by default

returns the actual state

=cut
sub force_delete
{
  my $self=shift();
  ${$self->{FORCE_DELETE}}=shift()?1:0 if(@_);
  return ${$self->{FORCE_DELETE}}?1:0;
}

=item show_deleted(<1/0>)

If i's set the nodes/ways/relation with "action='delete'" are listet as normal
returns the actual state

It's not set by default

=cut
sub show_deleted
{
  my $self=shift();
  ${$self->{IGNORE_DELETED}}=shift()?1:0 if(@_);
  return ${$self->{IGNORE_DELETED}}?1:0;
}

#=======================================================================

=item to_xml()

convert Element back to OSM-XML-Element

returns the a String of the OSM-XML-Tag

=cut
sub to_xml
{
  my $self=shift;
  my $in=shift || '';
  my $ret="$in<".$self->type().' ';

  while(my ($k,$v)=each(%{$self->{REF}}))
  {
    next unless defined $v;
    $ret.=qq!$k="$v" !;
  }

  if($self->has_tags())
  {
    $ret.=">\n";
    $ret.=$self->tags()->to_xml("$in  ");
    $ret.="$in</".$self->type().">\n";
  }
  else
  { $ret.="/>\n"; }

  return $ret;
}

1;

=back

=head1 See also

L<OSM::Tree::BASE>

L<OSM::Tree::Node>

L<OSM::Tree::Way>

L<OSM::Tree::Relation>

L<OSM::Tree::TagList>

L<OSM::Tree>

=cut
