package OSM::Tree::Polygon;
use strict;
use warnings;
use Math::Polygon;
use Scalar::Util qw(blessed looks_like_number);
use Math::Polygon::Calc;
use Math::Trig ':pi';

=encoding utf8

=head1 NAME

OSM::Tree::Polygon -- helperclass for OSM::Tree

=head1 SYNOPSIS

 #!/usr/bin/perl
 use strict;
 use warnings;
 use OSM::Tree;
 use OSM::Tree::Poligon;
 use Data::Dumper;

 my $osm=OSM::Tree->new('map.osm');

 my $poly1=OSM::Tree::Polygon->new( ($osm->way_list())[0] );
 print Dumper($poly1->center());

 my $poly2=($osm->way_list())[1]->polygon();
 print Dumper($poly2->center());

 my $poly3=OSM::Tree::Polygon->new(($osm->node_list())[0..5]);
 print Dumper($poly3->center());

=head1 DESCRIPTION

Class to handle ways and Nodelists as Polygons

Base Class is L<Math::Polygon> Look there for all other Methods

All listed methods understands OSM::Tree Objects.

=head1 AUTHOR

ToPeG

=head1 METHODS

=over 4

=cut

our $VERSION=0.001;

use parent 'Math::Polygon';

=item new(POINTS)

create new OSM::Tree::Polygon object;

The points can be Arrayrefs with two values or an OSM::Tree::Node or OSM::Tree::Way

=cut

sub new
{
  my $class = shift;
  my %options;

  if(ref($_[0]))
  { $options{points}=\@_; }
  else
  {
    pop(@_) if(@_%2);
    %options=@_;
  }

  ($options{points},undef)=__convert_pair(@{$options{points}}) if($options{points});

  $class->SUPER::new(%options);
}

=item equal((OTHER|ARRAY, [TOLERANCE])|POINTS)

Compare two polygons, on the level of points. When the polygons are
the same but rotated, this will return false. See L<same()|Math::Polygon/"Geometry">.

Function L<Math::Polygon::Calc::polygon_equal()|Math::Polygon::Calc/"FUNCTIONS">.

The points can be Arrayrefs with two values or an OSM::Tree::Node or OSM::Tree::Way

=cut
sub equal
{
  my $self  = shift;
  my ($other, $tolerance);
  if(@_ > 2 || ref($_[1]))
  { $other = __convert(@_); }
  else
  {
    my $ref=shift;
    ($other,undef)= __convert_pair(ref($ref) eq 'ARRAY'?@$ref:$ref);
    $tolerance = shift;
  }
  polygon_equal(scalar($self->points), $other, $tolerance);
}

=item same((OTHER|ARRAY, [TOLERANCE])|POINTS)

Compare two polygons, where the polygons may be rotated wrt each
other. This is (much) slower than L<equal()|Math::Polygon/"Geometry">, but some algorithms
will cause un unpredictable rotation in the result.

Function L<Math::Polygon::Calc::polygon_same()|Math::Polygon::Calc/"FUNCTIONS">.

The points can be Arrayrefs with two values or an OSM::Tree::Node or OSM::Tree::Way

=cut
sub same
{
  my $self = shift;
  my ($other, $tolerance);
  if(@_ > 2 || ref($_[1]))
  { $other = __convert(@_); }
  else
  {
    my $ref=shift;
    ($other,undef) = __convert_pair(ref($ref) eq 'ARRAY'?@$ref:$ref);
    $tolerance = shift;
  }
  polygon_same(scalar($self->points), $other, $tolerance);
}

#-----------------------------------------------------------------------
#-----------------------------------------------------------------------
#-----------------------------------------------------------------------

=item contains(POINT)

Returns a truth value indicating whether the point is inside the polygon.
On the edge is inside.

The point can be an Arrayref with two values or an OSM::Tree::Node or OSM::Tree::Way

=cut
sub contains
{
  my ($self, $point) = @_;
  my ($points,undef) = __convert_pair($point);
  return 0 unless($points && @$points);
  for(@$points)
  { return 0 unless(__polygon_contains_point($_, $self->points)); }
  return 1
}


=item contains_all(POINTS)

Returns a truth value indicating whether all points are inside the polygon.
On the edge is inside.

The points can be an Arrayref with two values or an OSM::Tree::Node or OSM::Tree::Way

=cut
sub contains_all
{
  my $self=shift;
  my ($points,undef) = __convert_pair(@_);
  return undef unless(@$points);
  my @p=$self->points;
  for(@$points)
  { return 0 unless(__polygon_contains_point($_, @p)); }
  return 1;
}

=item contains_one(POINTS)

Returns a truth value indicating whether one of all points is inside the polygon.
On the edge is inside.

The points can be an Arrayref with two values or an OSM::Tree::Node or OSM::Tree::Way

=cut
sub contains_one
{
  my $self=shift;
  my ($points,undef) = __convert_pair(@_);
  return undef unless(@$points);
  my @p=$self->points;
  for(@$points)
  { return 1 if(__polygon_contains_point($_, @p)); }
  return 0;
}

=item contains_list(POINTS)

Returns a list of points, who are inside the Polygon.
On the edge is inside.

The points can be an Arrayref with two values or an OSM::Tree::Node or OSM::Tree::Way

=cut
sub contains_list
{
  my $self=shift;
  my @ret=();
  my ($points,$objects) = __convert_pair(@_);
  return undef unless(@$points);
  my @p=$self->points;
  for my $pos (0..$#$points)
  {
    my $point=$points->[$pos];
    if(__polygon_contains_point($point, @p))
    { push(@ret,$objects->[$pos]); }
  }
  return @ret;
}

=item force_close()

close open Polynome

=cut
sub force_close
{
  my $self=shift;
  return if($self->isClosed());
  my @p=$self->points;
  $self->{MP_points}  = [ @p,$p[0] ];
}

=item crossing((OTHER|ARRAY, [TOLERANCE])|POINTS)

Returns a list of points. The ponits reperents the location where the two Lines crossing

The points can be an Arrayref with two values or an OSM::Tree::Node or OSM::Tree::Way

=cut
sub crossing
{
  my $self=shift;
  my ($other, $tolerance);
  if(@_ > 2 || ref($_[1]))
  { $other = __convert(@_); }
  else
  {
    my $ref=shift;
    ($other,undef)= __convert_pair(ref($ref) eq 'ARRAY'?@$ref:$ref);
    $tolerance = shift;
  }

  $tolerance=0.00001 unless(defined($tolerance));

  return() unless($other);

  my @points;

  my $my_lpoint;
  for my $my_point ($self->points())
  {
    my $other_lpoint;
    if($my_lpoint)
    {
      my $other_lpoint;
      for my $other_point ($other->points())
      {
        if($other_lpoint)
        {
          my @p=__crossing_lines($my_lpoint,$my_point,$other_lpoint,$other_point,$tolerance);
          push(@points,\@p) if(@p);
        }
        $other_lpoint=$other_point;
      }
    }
    $my_lpoint=$my_point;
  }

  return  @points;
}


########################################################################

sub __crossing_lines
{
  my ($p1,$p2,$p3,$p4,$eps)=@_;

  $eps=0.00001 unless(defined($eps));

  my $denom  = ($p4->[1]-$p3->[1]) * ($p2->[0]-$p1->[0]) - ($p4->[0]-$p3->[0]) * ($p2->[1]-$p1->[1]);
  my $numera = ($p4->[0]-$p3->[0]) * ($p1->[1]-$p3->[1]) - ($p4->[1]-$p3->[1]) * ($p1->[0]-$p3->[0]);
  my $numerb = ($p2->[0]-$p1->[0]) * ($p1->[1]-$p3->[1]) - ($p2->[1]-$p1->[1]) * ($p1->[0]-$p3->[0]);


  return (
    ($p1->[0] + $p2->[0])/2,
    ($p1->[1] + $p2->[1])/2,
  ) if (abs($numera) < $eps && abs($numerb) < $eps && abs($denom) < $eps);

  return if (abs($denom) < $eps);

  my $mua = $numera / $denom;
  my $mub = $numerb / $denom;
  return if ($mua < 0 || $mua > 1 || $mub < 0 || $mub > 1);

  return (
      $p1->[0] + $mua * ($p2->[0] - $p1->[0]),
      $p1->[1] + $mua * ($p2->[1] - $p1->[1]),
    );
}

sub __convert_pair
{
  my @list=();
  my @objs=();
  for my $obj (@_)
  {
    if(blessed($obj))
    {
      if($obj->isa('OSM::Tree::Way'))
      {
        for my $node ($obj->node_list())
        {
          push(@list,[$node->coordinate()]);
          push(@objs,$node);
        }
      }
      elsif($obj->isa('OSM::Tree::Node'))
      {
        push(@list,[$obj->coordinate()]);
        push(@objs,$obj);
      }
      elsif($obj->isa('OSM::Tree::Relation'))
      {
        my ($ret,$reto)= __convert_pair(map{$_->child()}$obj->member_list());
        if($ret)
        {
          push(@list,@$ret);
          push(@objs,@$reto);
        }
      }
      elsif($obj->isa('Math::Polygon'))
      {
        push(@list,$obj->points());
        push(@objs,$obj->points());
      }
    }
    elsif(ref($obj) eq 'ARRAY' && @$obj==2 && looks_like_number($obj->[0]) && looks_like_number($obj->[1]))
    {
      push(@list,$obj);
      push(@objs,$obj);
    }
  }
  return \@list,\@objs;
}

sub __max { $_[0] > $_[1]? $_[0]: $_[1]; }
sub __min { $_[0] < $_[1]? $_[0]: $_[1]; }

sub __polygon_contains_point_count($@)
{
  my $point = shift;
  return 0 if @_ < 3;

  my ($x, $y) = @$point;
  my @points=@_;

  unless($points[0]->[0]==$points[-1]->[0] && $points[0]->[1]==$points[-1]->[1])
  { push(@points,[@{$points[0]}]); }

  my ($px1, $py1) = @{ (shift(@points)) };
  my $count=0;

  while(@points)
  {
    my ($px2, $py2) = @{ (shift(@points)) };
    return 1 if (
                  $y==$py1 && $py1==$py2 &&
                  $x >= __max($px1,$px2) &&
                  $x <= __min($px1,$px2)
                );
    if(
        $y >  __min($py1,$py2) &&
        $y <= __max($py1,$py2) &&
        $x <= __max($px1,$px2) &&
        $py1 != $py2
      )
    {
      my $xinters = ($y-$py1)*($px2-$px1) / ($py2-$py1)+$px1;
      $count++ if ($px1 == $px2 || $x <= $xinters);
    }
    ($px1, $py1)=($px2, $py2);
  }

  return $count % 2;
}

sub __polygon_contains_point_angle($@)
{
  my $point = shift;
  return 0 if @_ < 3;

  my ($x, $y) = @$point;
  my @points=@_;

  unless($points[0]->[0]==$points[-1]->[0] && $points[0]->[1]==$points[-1]->[1])
  { push(@points,[@{$points[0]}]); }

  my $angle=0;

  for(1..$#points)
  {
    my $px1 = $points[$_]->[0]   - $x;
    my $py1 = $points[$_]->[1]   - $y;
    my $px2 = $points[$_-1]->[0] - $x;
    my $py2 = $points[$_-1]->[1] - $y;
    my $dtheta = atan2($py2,$px2) - atan2($py1,$px1);

    $dtheta -= pi2 while ($dtheta > pi);
    $dtheta += pi2 while ($dtheta < - pi() );
    $angle += $dtheta;
  }

  return 0 if( abs($angle) < pi );
  return 1
}

*__polygon_contains_point=\&Math::Polygon::Calc::polygon_contains_point;

=item switch_polygon_contains_function(<'count'|'angle'|'origin'>)

Class Method. Sets the Value for all OSM::Tree::Polygon objects!

Set the algorithmus for detecting when an point is inside a polygon.

Default is the function Function L<Math::Polygon::Calc::polygon_contains_point|Math::Polygon::Calc/"FUNCTIONS">.

'origin' set Math::Polygon::Calc::polygon_contains_point

'count'  a reimplemetation of polygon_contains_point

'angle'  is an implementation is to compute the sum of the angles made
between the test point and each pair of points making up the polygon.
This algorithmus is used in many CAD-Applications.

=cut
sub switch_polygon_contains_function
{
  no warnings 'redefine';
  shift() if(ref($_[0]));
  my $type=shift || 'origin';
  if($type eq 'count')
  {
    *__polygon_contains_point=\&__polygon_contains_point_count;
    return;
  }

  if($type eq 'angle')
  {
    *__polygon_contains_point=\&__polygon_contains_point_angle;
    return;
  }

  *__polygon_contains_point=\&Math::Polygon::Calc::polygon_contains_point;
  return;
}

1;

=back

=head1 See also

L<OSM::Tree::BASE>

L<OSM::Tree::Node>

L<OSM::Tree::Way>

L<OSM::Tree::Relation>

L<OSM::Tree::TagList>

L<OSM::Tree>

=cut
