########################################################################
package OSM::Tree::Way;

=encoding utf8

=head1 NAME

OSM::Tree::Way -- OSM Way

=head1 SYNOPSIS


=head1 DESCRIPTION

the Way-Object
derived from L<OSM::Tree::BASE>

=head1 AUTHOR

ToPeG

=head1 METHODS

=over 4

=cut

use strict;
use warnings;
use base qw( OSM::Tree::BASE OSM::Tree::_add_del );
my $Polygon;

sub BEGIN
{ $Polygon=1 if( eval { require OSM::Tree::Polygon; } ); }

sub _new
{
  my $class=shift;
  my $self={};
  $self->{TREE}=shift;
  $self->{ID}=shift;
  $self->{IGNORE_DELETED}=shift // \do{my $t=0};
  $self->{FORCE_DELETE}=shift  // \do{my $t=0};
  return undef unless(exists($self->{TREE}->{way}->{$self->{ID}}));
  $self->{REF}=$self->{TREE}->{way}->{$self->{ID}};
  $self->{bbox}=undef;
  bless($self,$class);
  return $self;
}

=item center()

get the center of this way

returns a longitude,latitude pair

or undef,undef if the way has no Nodes

Example:

 my ($lon,$lat)= $way->center();

=cut
sub center
{
  my $self=shift;
  my $mx=undef;
  my $my=undef;
  my @list=$self->node_list();
  return (undef, undef) unless(@list);

  # ignore first Note of an closed Way;
  shift(@list)if(
                  $list[0]->id() eq $list[-1]->id() ||
                  (
                    $list[0]->longitude() == $list[-1]->longitude() &&
                    $list[0]->latitude()  == $list[-1]->latitude()
                  )
                );

  for my $p (@list)
  {
    $mx+=$p->longitude();
    $my+=$p->latitude();
  }
  $mx/=@list;
  $my/=@list;

  return ($mx,$my);
}

=item length()

get the lenth of this way

returns the length

Example:

 my $length= $way->length();

=cut
sub length
{
  my $self=shift;
  my $length=0;
  my ($mx,$my)=(undef,undef);
  for my $p ($self->node_list())
  {
    my ($nx,$ny)=$p->coordinate();
    if(defined($mx) && defined($my))
    {
      my $d_lat = ($ny - $my) * 111.11 ;
      my $d_lon = ($nx - $mx) * cos ( $my / 360 * 3.14 * 2 ) * 111.11;
      $length += sqrt ($d_lat*$d_lat+$d_lon*$d_lon);
    }
    $mx=$nx;
    $my=$ny;
  }
  return $length;
}

=item bbox()

get the Boundary box of this way

returns (lonmin,latmin,lonmax,latmax)

Example:

 my ($lonin,$latmin)= $way->bbox();

=cut
sub bbox
{
  my $self=shift;
  my $force=shift;

  # if no boundary box is defined create one
  if(!defined($self->{bbox}) || ref($self->{bbox}) ne 'ARRAY' || $force)
  {
    my ($lonmin,$latmin,$lonmax,$latmax)=(undef,undef,undef,undef);
    for my $nid ($self->node_id_list())
    {
      my $nref=$self->{TREE}->{node}->{$nid};
      my ($lon,$lat)=($nref->{lon},$nref->{lat});

      if(defined($lonmin))
      { $lonmin = $lon if($lonmin > $lon); }
      else
      { $lonmin = $lon; }

      if(defined($lonmax))
      { $lonmax = $lon if($lonmax < $lon); }
      else
      { $lonmax = $lon; }

      if(defined($latmin))
      { $latmin = $lat if($latmin > $lat); }
      else
      { $latmin = $lat; }

      if(defined($latmax))
      { $latmax = $lat if($latmax < $lat); }
      else
      { $latmax = $lat; }
    }
    $self->{bbox}=[$latmin,$lonmin,$latmax,$lonmax];
  }

  return @{$self->{bbox}};
}

#=======================================================================

=item node(ID)

get a node-object corresponding to the ID and this way

see L<OSM::Tree::Node> for more

return a node-object
or undef if the ID not exists in this way

=cut
sub node
{
  my $self=shift;
  my $nodeid=shift;
  if(exists($self->{TREE}->{node}->{$nodeid}))
  { return OSM::Tree::Node->_new($self->{TREE},$nodeid,$self->{IGNORE_DELETED},$self->{FORCE_DELETE}); }
  return wantarray?():undef;
}

=item node_add([NodeID | <NODE> | <HASH>], <POS>)

add a node to the NodeList

see OSM::Tree::node_add() for more

return 1 on success

=cut
sub node_add
{
  my $self=shift;
  if(@_>0 && @_<=2)
  {
    my $node=shift;

    return 0 unless(defined($node));

    if($node!~/\D/)
    {
      my $pos=shift;
      return $self->node_id_add($node,$pos);
    }

    if(ref($node) eq 'HASH')
    {
      my $pos=shift;
      my $n=$self->_node_add($node);
      return 0 unless($n);
      return $self->node_id_add($n->id(),$pos);
    }

    if(ref($node) && $node->isa('OSM::Tree::Node'))
    {
      my $pos=shift;
      return $self->node_id_add($node->id(),$pos);
    }
  }

  if(@_)
  {
    pop(@_) if(@_%2);
    my %opts=@_;
    my $pos=$opts{pos};
    my $node=$self->_node_add(%opts);
    return 0 unless($node);
    return $self->node_id_add($node->id(),$pos);
  }

  return 0;
}

=item node_add_append(NodeID | <NODE> | <HASH>)

add a node to the NodeList

see OSM::Tree::node_add() for more

return 1 on success

=cut
sub node_add_append
{
  my $self=shift;
  if(@_>0 && @_< 2)
  {
    my $node=shift;
    return $self->node_add($node,@{shift->{REF}->{nd}}+0);
  }

  if(@_)
  {
    pop(@_) if(@_%2);
    my %opts=@_;
    return $self->node_add(\%opts,@{shift->{REF}->{nd}}+0);
  }

  return 0;
}

=item node_add_after([NodeID | <NODE> | <HASH>], <node>)

add a node to the NodeList after the defined Node or NodeId

see OSM::Tree::node_add() for more

return 1 on success

=cut
sub node_add_after
{
  my $self=shift;
  if(@_>0 && @_<=2)
  {
    my $node=shift;
    my $after=shift;

    return 0 unless(defined($node));
    return 0 unless(defined($after));

    if($after!~/\D/)
    {
      my $pos=0;
      my @l=$self->node_id_list();
      for (@l)
      {
        last if($_ == $after);
        $pos++;
      }
      $pos++;
      $pos=$#l if($pos>$#l);
      return $self->node_add($node,$pos);
    }

    if(ref($after) and $after->isa('OSM::Tree::Node'))
    {
      $after=$after->id();
      return $self->node_add_after($node,$after);
    }

    return 0;
  }

  if(@_)
  {
    pop(@_) if(@_%2);
    my %opts=@_;
    return $self->node_add_after(\%opts,$opts{after});
  }

  return 0;
}
=item node_add_bevor([NodeID | <NODE> | <HASH>], <node>)

add a node to the NodeList after the defined Node or NodeId

see OSM::Tree::node_add() for more

return 1 on success

=cut
sub node_add_bevor
{
  my $self=shift;
  if(@_>0 && @_<=2)
  {
    my $node=shift;
    my $bevor=shift;

    return 0 unless(defined($node));
    return 0 unless(defined($bevor));

    if($bevor!~/\D/)
    {
      my $pos=0;
      for ($self->node_id_list())
      {
        last if($_ == $bevor);
        $pos++;
      }
      $pos--;
      $pos=0 if($pos<0);
      return $self->node_add($node,$pos);
    }

    if(ref($bevor) and $bevor->isa('OSM::Tree::Node'))
    {
      $bevor=$bevor->id();
      return $self->node_add_bevor($node,$bevor);
    }

    return 0;
  }

  if(@_)
  {
    pop(@_) if(@_%2);
    my %opts=@_;
    return $self->node_add_bevor(\%opts,$opts{after});
  }

  return 0;
}


=item node_del(ID)

delete a node from the NodeList
return 1 on success

=cut
sub node_del
{
  my $self=shift;
  my $node=shift;
  return 0 unless($node && $node->isa('OSM::Tree::Node'));
  return $self->node_id_del($node->id());
}

=item node_list(<CODE>)

get a list of all node-objects corresponding to this way
if an coderefence is given it will use as filter.

see L<OSM::Tree::Node> for more

returns a list of node-objects
or an empty list if there are no nodes

=cut
sub node_list
{
  my $self=shift;
  my $filter=shift;
  my @list;
  if($filter)
  {
    if(ref($filter) eq 'CODE')
    {
      for($self->node_id_list())
      {
        my $node=$self->node($_);
        push(@list,$node) if($filter->($node))
      }
    }
    else
    { croak(q(Given filter is not a CodeRefence!)); }
  }
  else
  { @list=map{$self->node($_)}$self->node_id_list(); }
  return @list;
}

=item node_id_list()

get a list of all node-IDs corresponding to this way

returns a list of node-IDs
or an empty list if there are no node-IDs

=cut
sub node_id_list
{
  my $self=shift;
  $self->{REF}->{nd}=[] unless($self->{REF}->{nd});
  return @{$self->{REF}->{nd}};
}

=item node_id_add(ID,<POS>)

delete a node ID from the Node_list

return 1 on succcess

=cut
sub node_id_add
{
  my $self=shift;
  my $id=shift;
  my $pos=shift;
  return 0 if(!defined($id) || $id=~/\D/);
  if(defined($pos))
  {
    return 0 if($pos=~/\D/);
    return 0 if($pos>@{$self->{REF}->{nd}});
    splice(@{$self->{REF}->{nd}},$pos,0,$id);
    $self-> _set_modify();
    return 1;
  }
  push(@{$self->{REF}->{nd}},$id);
  $self-> _set_modify();
  return 1;
}

=item node_id_del(ID)

delete node by id

return 1 on succcess

=cut
sub node_id_del
{
  my $self=shift;
  my $id=shift;
  return 0 if(!defined($id) || $id=~/\D/);

  my $pos;
  for(0..$#{$self->{REF}->{nd}})
  {
    if($self->{REF}->{nd}->[$_] == $id)
    {
      $pos=$_;
      last;
    }
  }
  return 0 unless(defined($pos));

  splice(@{$self->{REF}->{nd}},$pos,1);
  $self-> _set_modify();
  return 1;
}

=item node_list_length()

get the length the list of all nodes

returns the length of the node-list

=cut
sub node_list_length
{ return scalar(shift->node_id_list()); }

=item node_exists(ID)

get the length the list of all nodes

returns the length of the node-list

=cut
sub node_exists
{
  my $self=shift;
  my $nid=shift;
  return 0 unless(defined($nid));
  if(ref($nid) && ref($nid) eq 'OSM::Tree::Node')
  { $nid=$nid->id(); }
  return 0 if($nid=~/\D/);
  for my $id ($self->node_id_list())
  { return 1 if($nid==$id); }
  return 0;
}

=item polygon(<OPTIONS>)

make OSM::Tree::Polygon Object from this way

returns undef or an OSM::Tree::Polygon Object

if OSM::Tree::Polygon is not loadable
(As example Math::Polygon is not einstalled )
this Method returns undef.

see: L<OSM::Tree::Polygon>


=cut
sub polygon
{
  unless($Polygon)
  {
    carp("Math::Polygon or OSM::Tree::Polygon not installed");
    return undef;
  }
  return OSM::Tree::Polygon->new(points=>[shift(@_)],@_);
}

#-----------------------------------------------------------------------
# overwrite

sub to_xml
{
  my $self=shift;
  my $in=shift || '';
  my $ref=$self->{REF};

  $ref->{visible} ||= 'true';
  my @vals;
  for my $name (qw(id version timestamp uid user visible changeset))
  { push(@vals,qq($name="$ref->{$name}")) if(defined($ref->{$name})); }

  my $txt=join(' ',@vals);

  if(exists($ref->{action}) && $ref->{action})
  { $txt.=qq( action="$ref->{action}"); }

  my $ret=$in.sprintf(q!<way %s !,$txt);
  $ret  .=qq!$in  <nd ref="$_"/>\n! for($self->node_id_list());
  #$ret  .=qq!$in  <nd ref="!.$_->id().qq!"/>\n! for($self->node_list());
  $ret  .=$self->tags()->to_xml("$in  ") if($self->has_tags());
  $ret  .="$in</way>\n";
  return $ret;
}

1;

=back

=head1 See also

L<OSM::Tree::BASE>

L<OSM::Tree::Node>

L<OSM::Tree::Way>

L<OSM::Tree::Relation>

L<OSM::Tree::TagList>

L<OSM::Tree::Polygon>

L<OSM::Tree>

=cut

