use strict;
use warnings;
=encoding utf8

=head1 NAME

OSM::Tree::_add_del -- helper package to add and Delete Elememts

=head1 SYNOPSIS

=head1 DESCRIPTION

helper Package to add Nodes Way Relations

=head1 AUTHOR

ToPeG

=head1 USING

L<OSM::Tree::Node>

L<OSM::Tree::Way>

L<OSM::Tree::Relation>

=head1 METHODS

=over 4

=cut

package OSM::Tree::_add_del;

=item _node_add(<HASH>)

add node-object

keys are:

=over

=item * id

the unique node ID

=item * lon

longitude

=item * lat

latitude

=item * user | uid

username or user id

=item * changeset

ChangeSet

=back

optional

=over

=item * visible

Is this an visible node? Default: "yes"

=item * version

Version of the Node Default: 1

=item * timestamp

Timestamp of change/creation dafault time()

=item * tags
hash of key/value pairs

=back

returns the an node-object or undef

see L<OSM::Tree::Node> and OSM-XML spec for more

=cut
sub _node_add
{
  my $self=shift;
  my $opts=$_[0] if($_[0] && ref($_[0]) eq 'HASH');
  my $vals={};

  unless($opts)
  {
    pop(@_) if(@_%2);
    $opts={@_};
  }

  if(!defined($opts->{id}) || $opts->{id}=~/\D/)
  {
    $self->{ERROR}='No valid Node ID (id)';
    return undef;
  }

  if(exists($self->{TREE}->{node}->{$opts->{id}}))
  {
    $self->{ERROR}='Node ID already exists (id)';
    return undef;
  }
  $vals->{id}=$opts->{id};

  if(!defined($opts->{lon}) || $opts->{lon}!~/[^\d.]/)
  {
    $self->{ERROR}='No valid Longitude (lon)';
    return undef;
  }
  $vals->{lon}=$opts->{lon};

  if(!defined($opts->{lat}) || $opts->{lat}!~/[^\d.]/)
  {
    $self->{ERROR}='No valid Longitude (lat)';
    return undef;
  }
  $vals->{lat}=$opts->{lat};

  if(!defined($opts->{user}) || !defined($opts->{uid}) || $opts->{uid}=~/\D/)
  {
    $self->{ERROR}='No valid user or userid (user,uid)';
    return undef;
  }
  $vals->{user}=$opts->{user} if($opts->{user});
  $vals->{uid}=$opts->{uid} if($opts->{uid});

  if(exists($opts->{changeset}) && $opts->{changeset}!~/\D/)
  {
    $self->{ERROR}='No valid changeset (changeset)';
    return undef;
  }
  $vals->{changeset}=$opts->{changeset};

  $opts->{visible}='true' if(!defined($opts->{visible}));
  $opts->{version}=1 unless(exists($opts->{version}));
  if(!$opts->{timestamp})
  {
    my $time=POSIX::strftime('%Y-%m-%dT%H:%M:%SZ',time());
    $opts->{timestamp}=$time;
    $opts->{version}++;
  }
  $vals->{timestamp}=$opts->{timestamp};
  $vals->{version}=$opts->{version};
  $vals->{visible}=$opts->{visible};

  $self->{TREE}->{node}->{$opts->{id}}=$vals;

  my $node=OSM::Tree::Node->_new($self->{TREE},$opts->{id},$self->{SD},$self->{FORCE_DELETE});

  $node->_modify();
  if($node && $opts->{tags} && ref($opts->{tags}) eq 'HASH')
  {
    my $list=$node->tags();
    while(my @l = each(%{$opts->{tags}}))
    { $list->tag_add(@l); }
  }

  return $node;
}

=item _node_del(NODE)

detete a node
returns 1 on success

=cut
sub _node_del
{
  my $self=shift;
  my $node=shift;
  return 0 unless(!defined($node) && $node->isa('OSM::Tree::Node'));

  unless($self->force_delete())
  {
    $node->_set_action('delete');
    return 1;
  }

  if(eval{$self->can('node_id_del')})
  {
    if($self->node_id_del($node->id()))
    {
      $node->{TREE}={};
      return 1;
    }
  }
  elsif($self->_node_id_del($node->id()))
  {
    $node->{TREE}={};
    return 1;
  }
  return 0;
}

=item _node_id_del(ID)

detete a node
returns 1 on success

=cut
sub _node_id_del
{
  my $self=shift;
  my $id=shift;
  return 0 if(!defined($id) || $id=~/\D/);
  return 0 if(!exists($self->{TREE}->{node}->{$id}));

  unless($self->force_delete())
  {
    $self->{TREE}->{node}->{$id}->{action}='delete';
    return 1;
  }

  delete($self->{TREE}->{node}->{$id});
  return 1;
}

=item _way_add(<HASH>)

add way-object

keys are:

=over

=item * id

the unique node ID

=item * user | uid

username or user id

=item * changeset

ChangeSet

=back

optional

=over

=item * visible

Is this an visible node? Default: "yes"

=item * version

Version of the Node Default: 1

=item * timestamp

Timestamp of change/creation Default: time()

=item * tags

hash of key/value pairs

=item * nodes

List of Node-IDs

=back

returns the an node-object or undef

see L<OSM::Tree::way> and OSM-XML spec for more

=cut
sub _way_add
{
  my $self=shift;
  my $opts=$_[0] if($_[0] && ref($_[0]) eq 'HASH');
  my $vals={};

  unless($opts)
  {
    pop(@_) if(@_%2);
    $opts={@_};
  }
  my $id=delete($opts->{id});
  if(!defined($opts->{id}) || $opts->{id}=~/\D/)
  {
    $self->{ERROR}='No valid Way ID (id)';
    return undef;
  }
  if(exists($self->{TREE}->{way}->{$opts->{id}}))
  {
    $self->{ERROR}='Way ID already exists (id)';
    return undef;
  }
  $vals->{id}=$opts->{id};

  if(!defined($opts->{user}) || !defined($opts->{uid}) || $opts->{uid}=~/\D/)
  {
    $self->{ERROR}='No Way user or userid (user,uid)';
    return undef;
  }
  $vals->{user}=$opts->{user} if($opts->{user});
  $vals->{uid}=$opts->{uid} if($opts->{uid});

  if(exists($opts->{changeset}) && $opts->{changeset}!~/\D/)
  {
    $self->{ERROR}='No valid changeset (changeset)';
    return undef;
  }
  $vals->{changeset}=$opts->{changeset};

  $opts->{visible}='true' if(!defined($opts->{visible}));
  $opts->{version}=1 unless(exists($opts->{version}));
  if(!$opts->{timestamp})
  {
    my $time=POSIX::strftime('%Y-%m-%dT%H:%M:%SZ',time());
    $opts->{timestamp}=$time;
    $opts->{version}++;
  }
  $vals->{timestamp}=$opts->{timestamp};
  $vals->{version}=$opts->{version};
  $vals->{visible}=$opts->{visible};

  $self->{TREE}->{way}->{$opts->{id}}=$vals;

  my $way=OSM::Tree::Way->_new($self->{TREE},$opts->{id},$self->{IGNORE_DELETED},$self->{FORCE_DELETE});

  if($way)
  {
    my $list=$way->tags();
    $list->tag_add('action','modfy');
    if($opts->{tags} && ref($opts->{tags}) eq 'HASH')
    {
      while(my @l = each(%{$opts->{tags}}))
      { $list->tag_add(@l); }
    }

    if($opts->{nodes} && ref($opts->{nodes}) eq 'ARRAY')
    {
      for my $id (@{$opts->{nodes}})
      { $way->node_add($id) if($id!~/\D/); }
    }

  }

  return $way;
}

=item _way_del(WAY)

detete a node
returns 1 on success

=cut
sub _way_del
{
  my $self=shift;
  my $way=shift;
  return 0 unless(!defined($way) && $way->isa('OSM::Tree::Way'));

  unless($self->force_delete())
  {
    $way->_set_action('delete');
    return 1;
  }

  if(eval{$self->can('way_id_del')})
  {
    if($self->way_id_del($way->id()))
    {
      $way->{TREE}={};
      return 1;
    }
  }
  elsif($self->_way_id_del($way->id()))
  {
    $way->{TREE}={};
    return 1;
  }
  return 0;
}

=item _way_id_del(ID)

detete a way
returns 1 on success

=cut
sub _way_id_del
{
  my $self=shift;
  my $id=shift;
  return 0 if(!defined($id) || $id=~/\D/);
  return 0 if(!exists($self->{TREE}->{way}->{$id}));

  unless($self->force_delete())
  {
    $self->{TREE}->{way}->{$id}->{action}='delete';
    return 1;
  }

  delete($self->{TREE}->{way}->{$id});
  return 1;
}

=item _relation_add(<HASH>)

add relation-object

keys are:

=over

=item * id

the unique node ID

=item * user | uid

username or user id

=item * changeset

ChangeSet

=back

optional

=over

=item * visible

Is this an visible node? Default: "yes"

=item * version

Version of the Node Default: 1

=item * timestamp

Timestamp of change/creation Default: time()

=item * tags

hash of key/value pairs

=item * members

List of Hashes with ref type role see L<OSM::Tree::Member> for more

=back

returns the an node-object or undef

see L<OSM::Tree::way> and OSM-XML spec for more

=cut
sub _relation_add
{
  my $self=shift;
  my $opts=$_[0] if($_[0] && ref($_[0]) eq 'HASH');
  my $vals={};

  unless($opts)
  {
    pop(@_) if(@_%2);
    $opts={@_};
  }
  my $id=delete($opts->{id});
  if(!defined($opts->{id}) || $opts->{id}=~/\D/)
  {
    $self->{ERROR}='No valid Way ID (id)';
    return undef;
  }
  if(exists($self->{TREE}->{relation}->{$opts->{id}}))
  {
    $self->{ERROR}='Way ID already exists (id)';
    return undef;
  }
  $vals->{id}=$opts->{id};

  if(!defined($opts->{user}) || !defined($opts->{uid}) || $opts->{uid}=~/\D/)
  {
    $self->{ERROR}='No Way user or userid (user,uid)';
    return undef;
  }
  $vals->{user}=$opts->{user} if($opts->{user});
  $vals->{uid}=$opts->{uid} if($opts->{uid});

  if(exists($opts->{changeset}) && $opts->{changeset}!~/\D/)
  {
    $self->{ERROR}='No valid changeset (changeset)';
    return undef;
  }
  $vals->{changeset}=$opts->{changeset};

  $opts->{visible}='true' if(!defined($opts->{visible}));
  $opts->{version}=1 unless(exists($opts->{version}));
  if(!$opts->{timestamp})
  {
    my $time=POSIX::strftime('%Y-%m-%dT%H:%M:%SZ',time());
    $opts->{timestamp}=$time;
    $opts->{version}++;
  }
  $vals->{timestamp}=$opts->{timestamp};
  $vals->{version}=$opts->{version};
  $vals->{visible}=$opts->{visible};

  $self->{TREE}->{relation}->{$opts->{id}}=$vals;

  my $relation=OSM::Tree::Relation->_new($self->{TREE},$opts->{id},$self->{IGNORE_DELETED},$self->{FORCE_DELETE});

  if($relation)
  {
    my $list=$relation->tags();
    $list->tag_add('action','modfy');

    if($opts->{tags} && ref($opts->{tags}) eq 'HASH')
    {
      while(my @l = each(%{$opts->{tags}}))
      { $list->tag_add(@l); }
    }

    if($opts->{members} && ref($opts->{members}) eq 'ARRAY')
    { $relation->add_member($_) for (@{$opts->{members}}); }
  }

  return $relation;
}


=item relation_del(RELATION)

detete a relation
returns 1 on success

=cut
sub _relation_del
{
  my $self=shift;
  my $relation=shift;
  return 0 unless(!defined($relation) && $relation->isa('OSM::Tree::Relation'));

  unless($self->force_delete())
  {
    $relation->_set_action('delete');
    return 1;
  }

  if(eval{$self->can('relation_id_del')})
  {
    if($self->relation_id_del($relation->id()))
    {
      $relation->{TREE}={};
      return 1;
    }
  }
  elsif($self->_relation_id_del($relation->id()))
  {
    $relation->{TREE}={};
    return 1;
  }
  return 0;
}

=item relation_id_del(ID)

detete a relation
returns 1 on success;

=cut
sub _relation_id_del
{
  my $self=shift;
  my $id=shift;
  return 0 if(!defined($id) || $id=~/\D/);
  return 0 if(!exists($self->{TREE}->{relation}->{$id}));
  unless($self->force_delete())
  {
    $self->{TREE}->{relation}->{$id}->{action}='delete';
    return 1;
  }
  delete($self->{TREE}->{relation}->{$id});
  return 1;
}

1;


=back

=cut

=head1 BUGS

None known.

=head1 See also

L<OSM::Tree::BASE>

L<OSM::Tree::Node>

L<OSM::Tree::Way>

L<OSM::Tree::Relation>

L<OSM::Tree>

=cut
